/*
    Project: TCGINA
    Version: V1.23
    File   : kfiencrypt.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "kfipriv.h"

#ifndef KFI_DECRYPTION_ONLY
#include "random.h"

int KFI_Init(KFI_DATA *pd)
{
    if (!GetRandom(((unsigned char *)pd) +
      KFI_DATA_OFFSET, KFI_DATA_ENC_SIZE)) {
        return CRYPT_ERROR;
    }

    pd->dwSignature     = KFI_SIGNATURE;
    pd->wType           = 0;
    pd->wVersion        = 0;
    pd->chDriveLetter   = '\0';
    pd->dwMountMask     = 0;
    pd->dwMountOptions  = 0;
    pd->wKeyType        = 0;
    pd->wKeySize        = 0;
    pd->wPasswordType   = 0;
    pd->wPasswordSize   = 0;
    *pd->wszHostName    = L'\0';
    *pd->aszPassword    = '\0';
    *pd->wszDescription = L'\0';
    pd->dwReserved      = 0;
    memset(pd->rgbReserved, '\0', sizeof(pd->rgbReserved));

    if (pd->dwID == 0) {
        pd->dwID = 1;
    }

    return CRYPT_OK;
}

int KFI_UpdateSalt(KFI_DATA *pd)
{
    return GetRandom(pd->rgbSalt, sizeof(pd->rgbSalt)) ? CRYPT_OK : CRYPT_ERROR;
}

#ifdef _MSC_VER
#  pragma warning( push )
#  pragma warning( disable : 4701 )
#endif

int KFI_Encrypt(KFI_DATA *pdDest, const KFI_DATA *pdSrc,
  const KFI_ENCRYPTION_DATA *ped)
{
    int               iCipher1;
    int               iCipher2;
    int               iError;
    BYTE              *pbDest;
    BYTE              *pbDestStart;
    BYTE              *pbDestEnd;
    BYTE              *pbPrevDest;
    KFI_DATA *pd;
    struct {
        BYTE          rgbState[KFI_CIPHER_BLOCK_SIZE];
        symmetric_key skKey1;
        symmetric_key skKey2;
    } *tmp;

    if (ped->fValid) {
        if (register_cipher(&KFI_CIPHER1_DESC) == -1) {
            return CRYPT_INVALID_CIPHER;
        }
        if ((iCipher1 = find_cipher(KFI_CIPHER1_NAME)) == -1) {
            return CRYPT_INVALID_CIPHER;
        }
    }
    if (register_cipher(&KFI_CIPHER2_DESC) == -1) {
        return CRYPT_INVALID_CIPHER;
    }
    if ((iCipher2 = find_cipher(KFI_CIPHER2_NAME)) == -1) {
        return CRYPT_INVALID_CIPHER;
    }

    if (!sec_alloc(tmp)) {
        return CRYPT_MEM;
    }
    if (pdDest == pdSrc) {
        if ((pd = sec_malloc(sizeof(KFI_DATA))) == NULL) {
            sec_free(tmp);
            return CRYPT_MEM;
        }
    } else {
        pd = pdDest;
    }
    memcpy(pd, pdSrc, sizeof(KFI_DATA));

    /* Update checksum */
    pd->dwCRC = calc_crc32(((unsigned char*)pd) + KFI_MESSAGE_OFFSET,
      KFI_MESSAGE_SIZE);

    /* Encrypt data */
    if (ped->fValid) {
        if ((iError = cipher_descriptor[iCipher1].setup(ped->rgbKey1,
          sizeof(ped->rgbKey1), 0, &tmp->skKey1)) != CRYPT_OK) {
            goto error;
        }
        if ((iError = cipher_descriptor[iCipher2].setup(ped->rgbKey2,
          sizeof(ped->rgbKey2), 0, &tmp->skKey2)) != CRYPT_OK) {
            goto error;
        }
        pbDestStart = ((unsigned char *)pd) + KFI_DATA_OFFSET;
        pbDestEnd   = pbDestStart + KFI_DATA_ENC_SIZE;

        pbPrevDest  = NULL;
        pbDest = pbDestStart;
        while (pbDest != pbDestEnd) {
            KFI_BlockXor(pbDest, pbPrevDest ? pbPrevDest : ped->rgbIV);
            cipher_descriptor[iCipher1].ecb_encrypt(pbDest, pbDest,
              &tmp->skKey1);
            cipher_descriptor[iCipher2].ecb_encrypt(pbDest, pbDest,
              &tmp->skKey2);
            pbPrevDest = pbDest;
            pbDest += KFI_CIPHER_BLOCK_SIZE;
        }
        pbDest = pbDestStart;
        while (pbDest != pbDestEnd) {
            KFI_BlockXor(pbDest, ped->rgbWhiteningKey);
            pbDest += KFI_CIPHER_BLOCK_SIZE;
        }
    } else {
        if ((iError = cipher_descriptor[iCipher2].setup(pd->rgbSalt,
          KFI_CIPHER_KEY_SIZE, 0, &tmp->skKey2)) != CRYPT_OK) {
            goto error;
        }
        pbDestStart = ((unsigned char *)pd) + KFI_DATA_OFFSET;
        pbDestEnd   = pbDestStart + KFI_DATA_ENC_SIZE;

        pbPrevDest  = NULL;
        pbDest = pbDestStart;
        while (pbDest != pbDestEnd) {
            KFI_BlockXor(pbDest, pbPrevDest ? pbPrevDest :
              (pd->rgbSalt+KFI_CIPHER_KEY_SIZE));
            cipher_descriptor[iCipher2].ecb_encrypt(pbDest, pbDest,
              &tmp->skKey2);
            pbPrevDest = pbDest;
            pbDest += KFI_CIPHER_BLOCK_SIZE;
        }
    }

    if (pdDest == pdSrc) {
        memcpy(pdDest, pd, sizeof(KFI_DATA));
    }

error:
    if (pdDest == pdSrc) {
        sec_free(pd);
    }
    sec_free(tmp);

    return iError;
}

#ifdef _MSC_VER
#  pragma warning( pop )
#endif

#endif
